@@ -128,6 +128,8 @@ static int mainqueue(int listener) {
 			if( c->client_fd == -1 && c->server_fd == -1 ) {
 				if( c == connections ) {
 					free_usedextensions(c->usedextensions);
+					free_unknownextensions(c->unknownextensions);
+					free_unknownextensions(c->waiting);
 					free(c->from);
 					connections = c->next;
 					free(c);
@@ -309,7 +311,6 @@ static int mainqueue(int listener) {
 		}
 
 	}
-	
 	return EXIT_SUCCESS;
 }
 
